"""
    land_use_perc(ws_host_port::String, selection::Dict) -> DataFrame

Retrieve land use percentages for specified geographic areas and year using a web service.

This function calculates land use percentages for specified geographic areas based on the provided shape data and year. It communicates with a web service to retrieve land use statistics for each area and returns the results as a DataFrame.

# Arguments
- `ws_host_port::String`: The host and port of the web service in the format "http://host:port".
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id`, `shape_obj`, and `year`.

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (geographic ID), `water_perc` (percentage of water), `tree_perc` (percentage of trees), `grass_perc` (percentage of grass), `flooded_perc` (percentage of flooded land), `crop_perc` (percentage of crops), `scrub_perc` (percentage of scrub land), `built_perc` (percentage of built land), `bare_perc` (percentage of bare land), `snow_perc` (percentage of land covered with snow), and `cloud_perc` (percentage of land cloud-covered).

# Example
> ws_host_port = "http://localhost:5000"
> selection = Dict("geo_id" => ["geo1", "geo2"], "shape_obj" => ["WKT1", "WKT2"], "year" => 2023)
> df = land_use_perc(ws_host_port, selection)

This example calculates land use percentages for two geographic areas using the provided shape data and year.
"""
function land_use_perc(ws_host_port::String, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    year = selection["year"]
    # Arrays to store column data
    geo_ids = String[]
    water = Float64[]
    tree = Float64[]
    grass = Float64[]
    flood = Float64[]
    crop = Float64[]
    scrub = Float64[]
    built = Float64[]
    bare = Float64[]
    snow = Float64[]
    cloud = Float64[]

    for (id, shape) in zip(geo_id, shape_obj)
        json_payload = JSON.json(Dict("polygon_wkt" => shape))
        url = "$ws_host_port/band_statistics/lc$year"
        headers = Dict("Content-Type" => "application/json")
        response = HTTP.post(url, headers, json_payload)
        data = JSON.parse(String(response.body))

        total_points = data["band_0"]["total_points"]
        categories = data["band_0"]["categories"]

        # Initialize all categories to zero
        category_values = Dict(1 => 0.0, 2 => 0.0, 3 => 0.0, 4 => 0.0, 5 => 0.0, 6 => 0.0, 7 => 0.0, 8 => 0.0, 9 => 0.0, 10 => 0.0, 11 => 0.0)

        # Update with actual values and calculate percentages
        for (cat, points) in categories
            cat = parse(Int, cat)  # Convert string key to integer
            if cat in keys(category_values)
                category_values[cat] = points / total_points
            end
        end

        push!(geo_ids, id)
        push!(water, category_values[1])
        push!(tree, category_values[2])
        push!(grass, category_values[3])
        push!(flood, category_values[4])
        push!(crop, category_values[5])
        push!(scrub, category_values[6]+category_values[11])
        push!(built, category_values[7])
        push!(bare, category_values[8])
        push!(snow, category_values[9])
        push!(cloud, category_values[10])
    end
    
    # object to be returned
    dftemp = DataFrame(
        geo_id = geo_ids,
        water_perc = water,
        tree_perc = tree,
        grass_perc = grass,
        flooded_perc = flood, 
        crop_perc = crop, 
        scrub_perc = scrub, 
        built_perc = built, 
        bare_perc = bare, 
        snow_perc = snow, 
        cloud_perc = cloud 
    )
    return dftemp
end

"""
get_land_use_data(
    service_query::String,
    selection::Dict,
    system::OrchestratorRegistry.ServiceRegistry
) -> DataFrame

Retrieves land use data from a specified service query and selection criteria.

This function dynamically connects to a service retrieved from the provided `ServiceRegistry`, fetches land use data based on the specified `service_query`, and applies the given selection criteria. The retrieved data includes various land use percentages, such as water, trees, grass, and more.

# Arguments
- `service_query::String`: The name of the service query to use for data retrieval, such as `'land_use_perc'`.
- `selection::Dict`: A dictionary specifying the selection criteria, including:
  - `"geo_id"`: A list of geographic IDs to filter by.
  - `"shape_obj"`: A list of WKT (Well-Known Text) representations of shapes to query.
  - `"year"`: The year to filter the data.
- `system::OrchestratorRegistry.ServiceRegistry`: A service registry used to retrieve the `FACT_lulc` service for connection details.

# Returns
- `DataFrame`: A DataFrame containing land use data with the following columns:
  - `geo_id`: Geographic ID of the region.
  - `water_perc`: Percentage of water coverage.
  - `tree_perc`: Percentage of tree coverage.
  - `grass_perc`: Percentage of grass coverage.
  - `flooded_perc`: Percentage of flooded land.
  - `crop_perc`: Percentage of crops.
  - `scrub_perc`: Percentage of scrub land.
  - `built_perc`: Percentage of built land.
  - `bare_perc`: Percentage of bare land.
  - `snow_perc`: Percentage of land covered with snow.
  - `cloud_perc`: Percentage of land covered by clouds.

# Behavior
- Retrieves the `FACT_lulc` service from the `ServiceRegistry` to configure the connection.
- Dynamically maps the `service_query` argument to the appropriate data-fetching function.
- Fetches and filters land use data based on the specified `selection` criteria.

# Raises
- `Error`: If the `service_query` argument does not match a supported service query.

# Example
Define the `selection` criteria and retrieve land use data:

```julia
# Define selection criteria
service_query = "land_use_perc"
selection = Dict(
    "geo_id" => ["geo1", "geo2"],
    "shape_obj" => ["WKT1", "WKT2"],
    "year" => 2023
)

# Create a ServiceRegistry and add the required service
system = ServiceRegistry()
add_service!(system, Service("FACT_lulc", "FACT_lulc_service", "http://localhost", 8080))

# Fetch land use data
df = get_land_use_data(service_query, selection, system)

println(df)
```

In this example, the function retrieves land use data for the specified `geo_id`, `shape_obj`, and year using the `'land_use_perc'` service query.
"""
function get_land_use_data(service_query::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    services = get_services_by_name(system, "FACT_lulc")
    if isempty(services)
        error("Service 'FACT_lulc' not found in the registry")
    end
    service = services[1]
    service_dict = Dict(
        "land_use_perc" => land_use_perc
    )
    df_template = DataFrame(
        geo_id = String[],  # Column for Origin Geo IDs
        water_perc = Float64[], # column for percentage of water
        tree_perc = Float64[], # column for percentage of trees
        grass_perc = Float64[], # column for percentage of grass
        flooded_perc = Float64[], # column for percentage of flooded land
        crop_perc = Float64[], # column for percentage of crops
        scrub_perc = Float64[], # column for percentage of scrub land
        built_perc = Float64[], # column for percentage of built land
        bare_perc = Float64[], # column for percentage of bare land
        snow_perc = Float64[], # column for percentage of land covered with snow
        cloud_perc = Float64[] # column for percentage of land cloud covered
    )

    # Check if the service is supported
    if !haskey(service_dict, service_query)
        error("Unsupported service: $service_query. Please check the available services: $(keys(service_dict)).")
    end

    # data retrieval
    df = get_results_from_service(service_query, df_template, service_dict, selection; host=service.host, port=service.port)
    return df
end
