# we do not parse EU grid info yet

"""
    eu_lau(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve European Union (EU) Local Administrative Units (LAU) population data for specified geographic areas and years.

This function retrieves population data for specified LAUs in the European Union for a given year. It communicates with a MySQL database connection to execute SQL queries and returns the results as a DataFrame.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying EU LAU population data.
- `selection::Dict`: A dictionary specifying the selection criteria including `year` (acceptable years: 1961, 71, 81, 91, 2001, 2011), and `geo_id` (LAU codes).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (LAU code), `region` (region identifier, always "EU"), and `population` representing the population count for the specified LAUs in the given year.

# Example
``` julia
conn = MySQL.Connection(user="user", password="password", host="localhost", db="eu_population")
selection = Dict("year" => 2011, "geo_id" => ["LAU1", "LAU2"])
df = eu_lau(conn, selection)
```

This example retrieves population data for two EU LAUs in the year 2011.
"""
function eu_lau(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary: acceptable years are 1961, 71, 81, 91, 2001, 2011
    year = selection["year"]
    geo_id = selection["geo_id"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for id in geo_id
        # Extract the first two characters of id in Julia
        id_short = id[1:2]
        # prepare query
        query = "SELECT POP_"*string(year)*"_01_01 as population FROM eu_lau WHERE eu_country = '"*id_short*"' AND CNTR_LAU_CODE = '"*id*"';"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    eu_nuts(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve European Union (EU) Nomenclature of Territorial Units for Statistics (NUTS) population data for specified geographic areas and years.

This function retrieves population data for specified NUTS regions in the European Union for a given year. It communicates with a MySQL database connection to execute SQL queries and returns the results as a DataFrame.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying EU NUTS population data.
- `selection::Dict`: A dictionary specifying the selection criteria including `year` (acceptable years: 2014-2022), and `geo_id` (NUTS codes).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (NUTS code), `region` (region identifier, always "EU"), and `population` representing the population count for the specified NUTS regions in the given year.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="eu_population")
> selection = Dict("year" => 2018, "geo_id" => ["NUTS1", "NUTS2"])
> df = eu_nuts(conn, selection)

This example retrieves population data for two EU NUTS regions in the year 2018.
"""
function eu_nuts(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary: acceptable years are 2014-2022
    year = selection["year"]
    geo_id = selection["geo_id"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for id in geo_id
        # Extract the first two characters of id in Julia
        id_short = id[1:2]
        # prepare query
        query = "SELECT Pop_M+Pop_F as population FROM eu_nuts WHERE eu_country = '"*id_short*"' AND AgeClass = 'TOTAL' AND GeoID = '"*id*"' AND Year = "*string(year)*";"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    us_block(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve United States (US) Census Block population data for specified geographic areas and years.

This function retrieves population data for specified US Census Blocks for a given year. It communicates with a MySQL database connection to execute SQL queries and returns the results as a DataFrame.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying US Census Block population data.
- `selection::Dict`: A dictionary specifying the selection criteria including `year` (acceptable years: 1961, 71, 81, 91, 2001, 2011, 2020), and `geo_id` (Block IDs).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (Block ID), `region` (region identifier, always "US"), and `population` representing the population count for the specified US Census Blocks in the given year.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="us_population")
> selection = Dict("year" => 2020, "geo_id" => ["010010201001000"])
> df = us_block(conn, selection)

This example retrieves population data for a US Census Block in the year 2020.
"""
function us_block(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary: acceptable years are 1961, 71, 81, 91, 2001, 2011
    year = selection["year"]
    geo_id = selection["geo_id"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    if year == 2020
        for id in geo_id
            state = parse(Int, id[1:2])
            county = parse(Int, id[3:5])
            tract = parse(Int, id[6:11])
            block = parse(Int, id[12:15])
            # prepare query
            query = "SELECT P1_001N as population FROM us_block_2020 WHERE state = "*string(state)*" AND county = "*string(county)*" AND tract = "*string(tract)*" AND block = "*string(block)*";"
            # create connection
            conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
            # execute
            result = DataFrame()
            try
                # Execute session variable adjustments
                result = query_connection(conn, query)
            catch ex
                @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
            end
            # close connection
            close_connection(conn)
            # Check if the result is empty
            if !isempty(result)
                # add geo_id col
                result[!, :geo_id] = fill(id, nrow(result))
                # copy results to be returned
                dftemp = vcat(dftemp, result)
            else
                println("No data returned from the query:")
                println(query)
            end
        end
    elseif year == 2010
        for id in geo_id
            # prepare query
            query = "SELECT pop10 as population FROM us_block_2010 WHERE blockid10 = '"*id*"';"
            # create connection
            conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
            # execute
            result = DataFrame()
            try
                # Execute session variable adjustments
                result = query_connection(conn, query)
            catch ex
                @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
            end
            # close connection
            close_connection(conn)
            # Check if the result is empty
            if !isempty(result)
                # add geo_id col
                result[!, :geo_id] = fill(id, nrow(result))
                # copy results to be returned
                dftemp = vcat(dftemp, result)
            else
                println("No data returned from the query:")
                println(query)
            end
        end
    end
    #dftemp[!, :region] = fill("US", nrow(dftemp))
    return dftemp
end

"""
    us_county(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve United States (US) County population data for specified geographic areas and years.

This function retrieves population data for specified US Counties for a given year. It communicates with a MySQL database connection to execute SQL queries and returns the results as a DataFrame.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying US County population data.
- `selection::Dict`: A dictionary specifying the selection criteria including `year` (acceptable years: 2014-2022), and `geo_id` (GeoIDs).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (GeoID), `region` (region identifier, always "US"), and `population` representing the population count for the specified US Counties in the given year.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="us_population")
> selection = Dict("year" => 2022, "geo_id" => ["01001"])
> df = us_county(conn, selection)

This example retrieves population data for a US County in the year 2022.
"""
function us_county(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary: acceptable years are 2014-2022
    year = selection["year"]
    geo_id = selection["geo_id"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for id in geo_id
        # prepare query
        query = "SELECT Pop_M+Pop_F as population FROM us_county WHERE AgeClass = 0 AND GeoID = '"*string(parse(Int, id))*"' AND Year = "*string(year)*";"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("US", nrow(dftemp))
    return dftemp
end

"""
get_annual_population(
    table::String,
    selection::Dict,
    system::OrchestratorRegistry.ServiceRegistry
) -> DataFrame

Retrieves annual population data for specified geographic areas and years.

This function dynamically connects to a service retrieved from the provided `ServiceRegistry`, fetches population data based on the specified `table`, and applies the given selection criteria. It supports querying population datasets for both EU and US geographic regions.

# Arguments
- `table::String`: The name of the table to query for population data, such as `'eu_lau'`, `'eu_nuts'`, `'us_block'`, or `'us_county'`.
- `selection::Dict`: A dictionary specifying the selection criteria, including:
  - `"year"`: The year to query.
  - `"geo_id"`: A list of geographic IDs to filter by.
- `system::OrchestratorRegistry.ServiceRegistry`: A service registry used to retrieve the `FACT_population` service for database connection details.

# Returns
- `DataFrame`: A DataFrame containing population data with the following columns:
  - `geo_id`: Geographic ID of the region.
  - `population`: Population count for the specified geographic areas in the given year.

# Behavior
- Retrieves the `FACT_population` service from the `ServiceRegistry` to configure the database connection.
- Dynamically maps the `table` argument to the appropriate data-fetching function.
- Fetches and filters population data based on the specified `selection` criteria.

# Raises
- `Error`: If the `table` argument does not match a supported table.

# Example
Define the `selection` criteria and retrieve annual population data:

```julia
# Define selection criteria
table = "eu_lau"
selection = Dict(
    "year" => 2011,
    "geo_id" => ["LAU1", "LAU2"]
)

# Create a ServiceRegistry and add the required service
system = ServiceRegistry()
add_service!(system, Service("FACT_population", "FACT_population_service", "http://localhost", 8080))

# Fetch population data
df = get_annual_population(table, selection, system)

println(df)
```

In this example, the function retrieves population data for two EU LAUs in the year 2011 using the `'eu_lau'` table.
"""
function get_annual_population(table::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    services = get_services_by_name(system, "FACT_population")
    if isempty(services)
        error("Service 'FACT_population' not found in the registry")
    end
    service = services[1]
    conn_dict = Dict(
        "host" => service.host,
        "port" => service.port,
        "database" => service.name,
        "user" => "root",
        "password" => "devops"
    )

    table_dict = Dict(
        "eu_lau" => eu_lau,
        "eu_nuts" => eu_nuts,
        "us_block" => us_block,
        "us_county" => us_county
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        population = Float64[] # column for average number of annual jobs
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end
    
    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 

    return df
end

