"""
    eu_ports(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve information about European Union (EU) ports, including the number of ports and total tonnage, for specified geographic areas.

This function communicates with a MySQL database connection to execute SQL queries and retrieve information about EU ports. It calculates the number of ports and total tonnage for the specified geographic areas based on their intersections with port shapes.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying EU port data.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id` (port IDs) and `shape_obj` (geometric shapes).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (port ID), `region` (region identifier, always "EU"), `ports` (number of ports), and `tonnage` (total tonnage in tonnage units) representing port information for the specified areas.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="eu_ports_db")
> selection = Dict("geo_id" => ["PORT1", "PORT2"], "shape_obj" => ["POINT(1 2)", "POINT(3 4)"])
> df = eu_ports(conn, selection)

This example retrieves information about two EU ports based on their specified port IDs and shapes.
"""
function eu_ports(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape) in zip(geo_id, shape_obj)
        # Extract the first two characters of id in Julia
        id_short = id[1:2]
        # prepare query
        query = "SELECT COUNT(*) as ports FROM eu_ports_2013_geo WHERE eu_country = '"*id_short*"' AND ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE);"
        query2 = "SELECT SUM(Q1+Q2+Q3+Q4) AS tonnage FROM eu_tonnage et JOIN eu_ports_2013_geo ep ON ep.port_id=et.PORT_ID WHERE ep.eu_country = '"*id_short*"' AND ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE);"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # execute
        result2 = DataFrame()
        try
            # Execute session variable adjustments
            result2 = query_connection(conn, query2)
        catch ex
            @error "Failed to set session variables for query $query2 with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        result = hcat(result, result2)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    us_ports(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve information about United States (US) ports, including the number of ports and total tonnage, for specified geographic areas.

This function communicates with a MySQL database connection to execute SQL queries and retrieve information about US ports. It calculates the number of ports and total tonnage for the specified geographic areas based on their intersections with port shapes.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying US port data.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id` (port IDs) and `shape_obj` (geometric shapes).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (port ID), `region` (region identifier, always "US"), `ports` (number of ports), and `tonnage` (total tonnage in tonnage units) representing port information for the specified areas.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="us_ports_db")
> selection = Dict("geo_id" => ["PORT1", "PORT2"], "shape_obj" => ["POINT(1 2)", "POINT(3 4)"])
> df = us_ports(conn, selection)

This example retrieves information about two US ports based on their specified port IDs and shapes.
"""
function us_ports(conn_dict::Dict, selection::Dict)
        # getting valuable info from dictionary: acceptable years are 1961, 71, 81, 91, 2001, 2011
        geo_id = selection["geo_id"]
        shape_obj = selection["shape_obj"]
        # object to be returned
        dftemp = DataFrame()
        # iterate over geo objects
        for (id, shape) in zip(geo_id, shape_obj)
            # prepare query
            query = "SELECT COUNT(*) as ports, SUM(total) as tonnage FROM us_ports_geo WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE);"
            # create connection
            conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
            # execute
            result = DataFrame()
            try
                # Execute session variable adjustments
                result = query_connection(conn, query)
            catch ex
                @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
            end
            # close connection
            close_connection(conn)
            # Check if the result is empty
            if !isempty(result)
                # add geo_id col
                result[!, :geo_id] = fill(id, nrow(result))
                # copy results to be returned
                dftemp = vcat(dftemp, result)
            else
                println("No data returned from the query:")
                println(query)
            end
        end 
        #dftemp[!, :region] = fill("EU", nrow(dftemp))
        return dftemp
end

"""
get_ports_tonnage(
    table::String,
    selection::Dict,
    system::OrchestratorRegistry.ServiceRegistry
) -> DataFrame

Retrieves port information, including the number of ports and total tonnage, for specified geographic areas.

This function dynamically connects to a service retrieved from the provided `ServiceRegistry`, fetches port data based on the specified `table`, and applies the given selection criteria. It supports querying datasets for both EU and US ports.

# Arguments
- `table::String`: The name of the table to query for port information, such as `'eu_ports'` or `'us_ports'`.
- `selection::Dict`: A dictionary specifying the selection criteria, including:
  - `"geo_id"`: A list of geographic IDs to filter by.
  - `"shape_obj"`: A list of WKT (Well-Known Text) representations of shapes to query.
- `system::OrchestratorRegistry.ServiceRegistry`: A service registry used to retrieve the `FACT_ports` service for database connection details.

# Returns
- `DataFrame`: A DataFrame containing port information with the following columns:
  - `geo_id`: Geographic ID of the region.
  - `ports`: Number of ports in the specified geographic area.
  - `tonnage`: Total tonnage of goods handled by the ports (in tonnage units).

# Behavior
- Retrieves the `FACT_ports` service from the `ServiceRegistry` to configure the database connection.
- Dynamically maps the `table` argument to the appropriate data-fetching function.
- Fetches and filters port information based on the specified `selection` criteria.

# Raises
- `Error`: If the `table` argument does not match a supported table.

# Example
Define the `selection` criteria and retrieve port information:

```julia
# Define selection criteria
table = "eu_ports"
selection = Dict(
    "geo_id" => ["PORT1", "PORT2"],
    "shape_obj" => ["POINT(1 2)", "POINT(3 4)"]
)

# Create a ServiceRegistry and add the required service
system = ServiceRegistry()
add_service!(system, Service("FACT_ports", "FACT_ports_service", "http://localhost", 8080))

# Fetch port information
df = get_ports_tonnage(table, selection, system)

println(df)
```

In this example, the function retrieves port information for two EU ports based on their geographic IDs and shapes using the `'eu_ports'` table.
"""
function get_ports_tonnage(table::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    services = get_services_by_name(system, "FACT_ports")
    if isempty(services)
        error("Service 'FACT_ports' not found in the registry")
    end
    service = services[1]
    conn_dict = Dict(
        "host" => service.host,
        "port" => service.port,
        "database" => service.name,
        "user" => "root",
        "password" => "devops"
    )
    table_dict = Dict(
        "eu_ports" => eu_ports,
        "us_ports" => us_ports
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        ports = Int64[], # column for number of stations
        tonnage = Float64[] # column to total length of lines
    )
    
    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end

    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 

    return df
end

