"""
    eu_rail(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve information about railway stations and lines in the European Union (EU) for specified geographic areas.

This function communicates with a MySQL database connection to execute SQL queries and retrieve information about railway stations and lines in the EU. It calculates the number of railway stations and the total length of railway lines for the specified geographic areas based on their intersections with station and line shapes.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying EU railway data.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id` (station IDs) and `shape_obj` (geometric shapes).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (station ID), `region` (region identifier, always "EU"), `stations` (number of railway stations), and `lines` (total length of railway lines in meters) representing railway station and line information for the specified areas.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="eu_rail_db")
> selection = Dict("geo_id" => ["STATION1", "STATION2"], "shape_obj" => ["POINT(1 2)", "POINT(3 4)"])
> df = eu_rail(conn, selection)

This example retrieves information about two EU railway stations and their associated railway lines based on specified station IDs and shapes.
"""
function eu_rail(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape) in zip(geo_id, shape_obj)
        # prepare query
        query = "SELECT COUNT(*) as stations FROM eu_stations WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE);"
        query2 = "SELECT SUM(ST_Length(ST_Intersection(SHAPE, ST_GeomFromText('"*shape*"')))) AS 'lines' FROM eu_lines WHERE ST_Intersects(SHAPE, ST_GeomFromText('"*shape*"'));"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # execute
        result2 = DataFrame()
        try
            # Execute session variable adjustments
            result2 = query_connection(conn, query2)
        catch ex
            @error "Failed to set session variables for query $query2 with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        result = hcat(result, result2)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
    us_rail(conn::MySQL.Connection, selection::Dict) -> DataFrame

Retrieve information about railway stations and lines in the United States (US) for specified geographic areas.

This function communicates with a MySQL database connection to execute SQL queries and retrieve information about railway stations and lines in the US. It calculates the number of railway stations and the total length of railway lines for the specified geographic areas based on their intersections with station and line shapes.

# Arguments
- `conn::MySQL.Connection`: The MySQL database connection to use for querying US railway data.
- `selection::Dict`: A dictionary specifying the selection criteria including `geo_id` (station IDs) and `shape_obj` (geometric shapes).

# Returns
- `DataFrame`: A DataFrame containing columns for `geo_id` (station ID), `region` (region identifier, always "US"), `stations` (number of railway stations), and `lines` (total length of railway lines in meters) representing railway station and line information for the specified areas.

# Example
> conn = MySQL.Connection(user="user", password="password", host="localhost", db="us_rail_db")
> selection = Dict("geo_id" => ["STATION1", "STATION2"], "shape_obj" => ["POINT(1 2)", "POINT(3 4)"])
> df = us_rail(conn, selection)

This example retrieves information about two US railway stations and their associated railway lines based on specified station IDs and shapes.
"""
function us_rail(conn_dict::Dict, selection::Dict)
    # getting valuable info from dictionary
    geo_id = selection["geo_id"]
    shape_obj = selection["shape_obj"]
    # object to be returned
    dftemp = DataFrame()
    # iterate over geo objects
    for (id, shape) in zip(geo_id, shape_obj)
        # prepare query
        query = "SELECT COUNT(*) as stations FROM na_stations WHERE ST_Intersects(ST_GeomFromText('"*shape*"'), SHAPE);"
        query2 = "SELECT SUM(ST_Length(ST_Intersection(SHAPE, ST_GeomFromText('"*shape*"')))) AS 'lines' FROM na_lines WHERE ST_Intersects(SHAPE, ST_GeomFromText('"*shape*"'));"
        # create connection
        conn = establish_connection(conn_dict["host"], conn_dict["user"], conn_dict["password"], conn_dict["database"]; port=conn_dict["port"])
        # execute
        result = DataFrame()
        try
            # Execute session variable adjustments
            result = query_connection(conn, query)
        catch ex
            @error "Failed to set session variables for query $query with expection " exception=(ex, catch_backtrace())
        end
        # execute
        result2 = DataFrame()
        try
            # Execute session variable adjustments
            result2 = query_connection(conn, query2)
        catch ex
            @error "Failed to set session variables for query $query2 with expection " exception=(ex, catch_backtrace())
        end
        # close connection
        close_connection(conn)
        result = hcat(result, result2)
        # Check if the result is empty
        if !isempty(result)
            # add geo_id col
            result[!, :geo_id] = fill(id, nrow(result))
            # copy results to be returned
            dftemp = vcat(dftemp, result)
        else
            println("No data returned from the query:")
            println(query)
        end
    end 
    #dftemp[!, :region] = fill("EU", nrow(dftemp))
    return dftemp
end

"""
get_rail_stations_lines(
    table::String,
    selection::Dict,
    system::OrchestratorRegistry.ServiceRegistry
) -> DataFrame

Retrieves railway station and line information for specified geographic areas.

This function dynamically connects to a service retrieved from the provided `ServiceRegistry`, fetches railway station and line data based on the specified `table`, and applies the given selection criteria. It supports querying datasets for both EU and US railway infrastructure.

# Arguments
- `table::String`: The name of the table to query for railway station and line information, such as `'eu_rail'` or `'us_rail'`.
- `selection::Dict`: A dictionary specifying the selection criteria, including:
  - `"geo_id"`: A list of geographic IDs to filter by.
  - `"shape_obj"`: A list of WKT (Well-Known Text) representations of shapes to query.
- `system::OrchestratorRegistry.ServiceRegistry`: A service registry used to retrieve the `FACT_rail` service for database connection details.

# Returns
- `DataFrame`: A DataFrame containing railway station and line information with the following columns:
  - `geo_id`: Geographic ID of the region.
  - `stations`: Number of railway stations in the specified geographic area.
  - `lines`: Total length of railway lines (in meters).

# Behavior
- Retrieves the `FACT_rail` service from the `ServiceRegistry` to configure the database connection.
- Dynamically maps the `table` argument to the appropriate data-fetching function.
- Fetches and filters railway infrastructure data based on the specified `selection` criteria.

# Raises
- `Error`: If the `table` argument does not match a supported table.

# Example
Define the `selection` criteria and retrieve railway infrastructure information:

```julia
# Define selection criteria
table = "eu_rail"
selection = Dict(
    "geo_id" => ["STATION1", "STATION2"],
    "shape_obj" => ["POINT(1 2)", "POINT(3 4)"]
)

# Create a ServiceRegistry and add the required service
system = ServiceRegistry()
add_service!(system, Service("FACT_rail", "FACT_rail_service", "http://localhost", 8080))

# Fetch railway station and line information
df = get_rail_stations_lines(table, selection, system)

println(df)
```

In this example, the function retrieves railway station and line information for two EU railway stations based on their geographic IDs and shapes using the `'eu_rail'` table.
"""
function get_rail_stations_lines(table::String, selection::Dict, system::OrchestratorRegistry.ServiceRegistry)
    services = get_services_by_name(system, "FACT_rail")
    if isempty(services)
        error("Service 'FACT_rail' not found in the registry")
    end
    service = services[1]
    conn_dict = Dict(
        "host" => service.host,
        "port" => service.port,
        "database" => service.name,
        "user" => "root",
        "password" => "devops"
    )

    table_dict = Dict(
        "eu_rail" => eu_rail,
        "us_rail" => us_rail
    )

    df_template = DataFrame(
        #region = String[], # column to store if EU or US
        geo_id = String[],  # Column for Geo IDs
        stations = Int64[], # column for number of stations
        lines = Float64[] # column to total length of lines
    )

    # Check if the table is supported
    if !haskey(table_dict, table)
        error("Unsupported table: $table. Please check the available tables: $(keys(table_dict)).")
    end    

    # data retrieval
    df = get_selected_objects_from_table(table, df_template, table_dict, selection, conn_dict) 

    return df
end

