"""
    interpret_scale(scale::String) -> Tuple

Interprets the `scale` string by splitting it into components based on specified separators.

This function splits the input `scale` string into parts using a regular expression that 
matches any non-word character (excluding underscores) or whitespace. It is designed to 
parse a string into distinct elements based on these separators.

# Arguments
- `scale::String`: The input string to be parsed.

# Returns
- `Tuple`: A tuple containing the separated parts of the input string.

# Example
```julia
interpret_scale("EU_2020 USA_2021")
# Returns: ("EU_2020", "USA_2021")
```
# Notes

The function uses a regular expression to identify separators, which include any
non-word character (like punctuation) and whitespace, but excludes underscores.
This is useful for parsing strings where different components are separated by
spaces, commas, or other non-word characters, while underscores are part of the
individual components.
"""
function interpret_scale(scale::String)
    # Define a regular expression for separators (excluding underscore)
    separator_regex = r"[^\w_]|[\s]"

    # Split the string based on the occurrence of the separator
    parts = split(scale, separator_regex)

    # Return a tuple containing all parts
    return Tuple(parts)
end

"""
    handle_simple_action(table::String, table_dict::Dict, conn::MySQL.Connection) -> DataFrame

Executes a specific action based on the given `table` name, using a dictionary of functions and a MySQL connection.

This function looks up a corresponding function in `table_dict` based on the `table` name and executes it 
using the provided MySQL connection. If no matching function is found in `table_dict` for the given `table` name, 
a default function is executed that returns a DataFrame with a message indicating no corresponding table was found.

# Arguments
- `table::String`: The name of the table for which an action needs to be executed.
- `table_dict::Dict`: A dictionary mapping table names to specific functions that handle actions for those tables.
- `conn::MySQL.Connection`: The MySQL connection to use for executing any database queries within the action functions.

# Returns
- `DataFrame`: A DataFrame containing the results of the action executed. If no action is found for the given `table`, 
  returns a DataFrame with a single column `Message` and a message indicating no corresponding table.

# Notes
- The functions in `table_dict` are expected to be of the form `func(conn::MySQL.Connection) -> DataFrame`.
- This allows for flexible and dynamic execution of table-specific actions based on the `table` name.
"""
function handle_simple_action(table::String, table_dict::Dict, conn_dict::Dict)
    # Default function returns an empty DataFrame with an appropriate message
    default_func = () -> DataFrame(Message=["No corresponding table: $table"])

    # Retrieve the function from the dictionary
    func = get(table_dict, table, default_func)

    # Call the function with the additional parameters
    return func(conn_dict)
end

"""
    get_all_objects_from_table(scale::String, df::DataFrame, table_dict::Dict; host::String="127.0.0.1") -> DataFrame

Fetches and compiles objects into a DataFrame based on a specified scale, using predefined actions for each table.

This function interprets a `scale` string to determine which tables to query. For each table identified in the 
`scale`, it executes a corresponding action from `table_dict` and appends the results to the provided DataFrame `df`. 
It ensures that the types of the columns in the appended data match the types in `df`.

# Arguments
- `scale::String`: A string that, when interpreted, defines the tables and/or actions to be queried.
- `df::DataFrame`: The DataFrame to which the results will be appended, defining the data types of contents.
- `table_dict::Dict`: A dictionary mapping table names to functions. Each function is responsible for fetching data 
  from the corresponding table and should return a DataFrame.
- `host::String`: (Optional) The host address for the MySQL database connection. Defaults to "127.0.0.1".

# Returns
- `DataFrame`: The original DataFrame `df` appended with the results from each table's action.

# Usage
The function is designed to work with a flexible scale definition. The `scale` string is interpreted to identify 
specific tables or actions. For each identified table, the function fetches data using the corresponding function in `table_dict`.

# Notes
- The function establishes and closes a MySQL database connection within its scope.
- It performs type checking and conversion to ensure that the data types of the columns in the result match those in the original `df`.
- Additional data type conversions should be added as necessary for other column types.
"""
function get_all_objects_from_table(scale::String, df::DataFrame, table_dict::Dict, conn_dict::Dict)
    # Interpret the scale using an external function
    interpreted_scale = interpret_scale(scale)
    
    for table in interpreted_scale
        result = handle_simple_action(String(table), table_dict, conn_dict)
        result = result[:, names(df)]
        # Check and convert each column in result to match the type in df
        for col in names(df)
            df_col_type = eltype(df[!, col])
            result_col_type = eltype(result[!, col])
    
            # Convert only if types mismatch
            if df_col_type != result_col_type
                if df_col_type == String
                    result[!, col] = string.(result[:, col])
                elseif df_col_type == Float64
                    result[!, col] = [tryparse(Float64, x) === nothing ? 0.0 : tryparse(Float64, x) for x in result[:, col]]
                elseif df_col_type == Int64
                    result[!, col] = [tryparse(Int64, x) === nothing ? 0 : tryparse(Int64, x) for x in result[:, col]]
                end
            end
        end
        df = vcat(df, result)
    end

    return df
end

"""
    handle_selection_action(table::String, table_dict::Dict, conn::MySQL.Connection, selection::Dict) -> DataFrame

Handle the selection action for a given table and return the appropriate DataFrame.

# Arguments
- `table::String`: The name of the table to handle.
- `table_dict::Dict`: A dictionary mapping table names to their corresponding functions.
- `conn::MySQL.Connection`: The MySQL database connection.
- `selection::Dict`: A dictionary containing selection parameters.

# Returns
- `DataFrame`: A DataFrame resulting from the called function for the given table. If the table is not found in `table_dict`, it returns a DataFrame with a message indicating no corresponding table.

# Description
This function looks up the table name in `table_dict` and retrieves a function associated with that table. It then calls this function, passing the MySQL connection and the selection dictionary as arguments. If the table name is not found in `table_dict`, it returns a default DataFrame with a message indicating that no corresponding table was found.

# Example
```julia
conn = MySQL.connect("host", "user", "password", "dbname")
table_dict = Dict("users" => users_func, "orders" => orders_func)
selection = Dict("id" => 123)
result = handle_selection_action("users", table_dict, conn, selection)
```
"""
function handle_selection_action(table::String, table_dict::Dict, conn_dict::Dict, selection::Dict)
    # Default function returns an empty DataFrame with an appropriate message
    default_func = () -> DataFrame(Message=["No corresponding table: $table"])

    # Retrieve the function from the dictionary
    func = get(table_dict, table, default_func)

    # Call the function with the additional parameters
    return func(conn_dict, selection)
end

"""
    get_selected_objects_from_table(table::String, df::DataFrame, table_dict::Dict, database::String, selection::Dict; host::String="127.0.0.1", port::Int=3306) -> DataFrame

Retrieve and append selected objects from a specified table in the database to a given DataFrame.

# Arguments
- `table::String`: The name of the table from which to retrieve objects.
- `df::DataFrame`: The DataFrame to which the results will be appended.
- `table_dict::Dict`: A dictionary mapping table names to their corresponding functions.
- `database::String`: The name of the database.
- `selection::Dict`: A dictionary containing selection parameters.
- `host::String` (optional): The database host address, default is "127.0.0.1".
- `port::Int` (optional): The port number for the database connection, default is 3306.

# Returns
- `DataFrame`: A DataFrame that combines the original `df` and the results retrieved from the database.

# Description
This function connects to a MySQL database and retrieves data based on the specified `selection` criteria from the `table`. The retrieved data is then processed to match the data types of the columns in the provided `df` DataFrame. If the types differ, conversion is attempted to align them with `df`. The results are appended to `df` and returned. The connection to the database is closed before returning the DataFrame.

# Example
```julia
df = DataFrame(id = Int[], name = String[])
table_dict = Dict("users" => users_func, "orders" => orders_func)
selection = Dict("id" => 123)
database = "my_database"
result_df = get_selected_objects_from_table("users", df, table_dict, database, selection)
```
"""
function get_selected_objects_from_table(table::String, df::DataFrame, table_dict::Dict, selection::Dict, conn_dict::Dict)
    
    # get results
    result = handle_selection_action(String(table), table_dict, conn_dict, selection)
    # start converting data types to fix mistakes in SQL types interpretation
    result = result[:, names(df)]
    # Check and convert each column in result to match the type in df
    for col in names(df)
        df_col_type = eltype(df[!, col])
        result_col_type = eltype(result[!, col])

        # Convert only if types mismatch
        if df_col_type != result_col_type
            if df_col_type == String
                result[!, col] = string.(result[:, col])
            elseif df_col_type == Float64
                result[!, col] = [tryparse(Float64, x) === nothing ? 0.0 : tryparse(Float64, x) for x in result[:, col]]
            elseif df_col_type == Int64
                result[!, col] = [tryparse(Int64, x) === nothing ? 0 : tryparse(Int64, x) for x in result[:, col]]
            end
        end
    end
    df = vcat(df, result)

    return df
end
