"""
    get_results_from_service(service_query::String, df::DataFrame, service_dict::Dict, selection::Dict; host::String="127.0.0.1", port::Int=80) -> DataFrame

Query a specified service and append the results to a given DataFrame after type conversion.

# Arguments
- `service_query::String`: The query string to identify which service to use.
- `df::DataFrame`: The DataFrame to which the results will be appended.
- `service_dict::Dict`: A dictionary mapping service queries to their corresponding functions.
- `selection::Dict`: A dictionary containing parameters for the service selection.
- `host::String` (optional): The hostname or IP address of the service, defaults to "127.0.0.1".
- `port::Int` (optional): The port number of the service, defaults to 80.

# Returns
- `DataFrame`: The original DataFrame combined with the results from the service query.

# Description
This function constructs a URL from the provided `host` and `port` and uses it to query a specified service. The service is determined by the `service_query` parameter and the corresponding function is fetched from `service_dict`. The `selection` dictionary is used as parameters for the query.

The function retrieves data and appends it to the provided DataFrame `df`. Before appending, it ensures that the data types of the retrieved results match those in `df`. If there's a mismatch, the function attempts to convert the types in the results to match those in `df`.

The function is particularly useful for dynamically querying different services and aggregating the results in a structured format. It's designed to handle common issues such as type mismatches often encountered when dealing with external data sources.

# Example
```julia
df = DataFrame(id = Int[], name = String[])
service_dict = Dict("getUserData" => userDataFunc, "getOrderData" => orderDataFunc)
selection = Dict("userId" => 123)
result_df = get_results_from_service("getUserData", df, service_dict, selection, host="api.example.com", port=8080)
'''
"""
function get_results_from_service(service_query::String, df::DataFrame, service_dict::Dict, selection::Dict; host::String="127.0.0.1", port::Int=80)
    ws_host_port = "http://$host:$port"
    # get results
    result = handle_selected_action(String(service_query), service_dict, ws_host_port, selection)
    # start converting data types to fix mistakes in SQL  types interpretation
    result = result[:, names(df)]
    # Check and convert each column in result to match the type in df
    for col in names(df)
        df_col_type = eltype(df[!, col])
        result_col_type = eltype(result[!, col])

        # Convert only if types mismatch
        if df_col_type != result_col_type
            if df_col_type == String
                result[!, col] = string.(result[:, col])
            elseif df_col_type == Float64
                result[!, col] = [tryparse(Float64, x) for x in result[:, col]]
            end
        end
    end
    df = vcat(df, result)
    return df
end
"""
    handle_selected_action(service_query::String, service_dict::Dict, ws_host_port::String, selection::Dict) -> DataFrame

Executes a selected action based on a given service query and returns the result as a DataFrame.

# Arguments
- `service_query::String`: The query string used to identify the specific action to be executed.
- `service_dict::Dict`: A dictionary mapping service queries to their corresponding functions.
- `ws_host_port::String`: The host and port of the web service as a string (e.g., "http://127.0.0.1:80").
- `selection::Dict`: A dictionary containing parameters for the action.

# Returns
- `DataFrame`: The result of the executed action. If the `service_query` does not correspond to any function in `service_dict`, a DataFrame with an error message is returned.

# Description
This function looks up the `service_query` in `service_dict` to find the corresponding function to execute. If the query matches a key in the dictionary, the associated function is called with `ws_host_port` and `selection` as arguments. If no matching key is found, a default function is called that returns a DataFrame with an error message indicating no corresponding web service query function was found.

This function is particularly useful for handling various service actions dynamically, allowing for flexible and scalable service interactions.

# Example
```julia
service_dict = Dict("getData" => getDataFunc, "postData" => postDataFunc)
selection = Dict("param1" => "value1", "param2" => "value2")
ws_host_port = "http://127.0.0.1:80"
result = handle_selected_action("getData", service_dict, ws_host_port, selection)
```
"""
function handle_selected_action(service_query::String, service_dict::Dict, ws_host_port::String, selection::Dict)
    # Default function returns an empty DataFrame with an appropriate message
    default_func = () -> DataFrame(Message=["No corresponding webservice query function: $service_query"])

    # Retrieve the function from the dictionary
    func = get(service_dict, service_query, default_func)

    # Call the function with the additional parameters
    return func(ws_host_port, selection)
end